\ Forth File System interface to the Unix operating system

\ Very-low-level interfaces to the Sun 4.2 BSD version of Unix
\ These routines may have to be rewritten for other flavors of
\ Unix, because the mechanism for doing system calls differs.

\ This version assumes that a C program is the loader for Forth.
\ The C program provides the Forth kernel with an array of entry-points
\ into C subroutines for performing the actual system calls.
\ This scheme should be reasonably compatible with nearly any Unix
\ implementation.  The only difference would be in the implementation of
\ "syscall", which has to look up the address of the actual system call
\ C routine in the system call table provided to it by the C program loader.
\ It then has to convert the stack arguments into the same form as is
\ expected by the C system call routines.  This obviously depends on the
\ details of the C calling sequence, but should not be too hard because
\ C compilers usually pass arguments on the stack.
\    Syscall is defined in the kernel, because it is needed for basics like
\ key and emit.

\ Offsets into the system-call vector.  These must agree with the order
\ of the system-call routines provided by the C program Forth loader.

\ Commented-out constants already defined in ../io.f
\  8 constant sys_open       12 constant sys_create
\ 16 constant sys_close      20 constant sys_read 
\ 24 constant sys_write
40 constant sys_lseek
44 constant sys_unlink

\ Constants for 4.2 BSD Unix seek system call options
0 constant u.l_set     1 constant u.l_incr  2 constant u.l_xtnd

\ Constants representing the allowable access modes for unix files
0 constant u_read-only   1 constant u_write-only  2 constant u_read-write

\ Efficient buffer size
td 512 bufsize !
th 1ff not constant u.recmsk

\ translation table between unix modes and forth file modes 
: constant-array ( -- name )
  create
  does> ( index -- value )
  swap /n * + @
;
constant-array >unixmode
   0 , u_read-only , u_write-only , u_read-write ,

\ protection to be assigned to newly-created files
\ read-write for owner and group, read only for other
variable file-protection

8 base !
664 file-protection !
decimal

create strbuf 100 allot
\ Packed forth string to C-style null-terminated string
: cstr ( forth_string --- d.c_string )
  count 2dup  strbuf swap cmove  ( forth-string-addr len )
  strbuf +  0 swap c!            ( forth-string-addr )
  drop
  strbuf n->a  ( relocate the address of the c-string if necessary )
;

: unixalign ( d.byteno --- d.aligned )
 ( rounds down to a block boundary )
  u.recmsk s->l land
;

\ C-style interfaces to Unix system calls already defined in ../io.f
\ : u_creat ( file-protection name -- unix-fd )
\   swap n->l nlswap cstr sys_create syscall retval >r  ldrop ldrop r>
\ ;  
\ : u_close ( unix_fd --- )
\   n->l sys_close syscall ldrop
\ ;
\ : u_open ( unixmode name --- unix-fd )
\   swap n->l nlswap cstr sys_open syscall retval >r ldrop ldrop r>
\ ;
: u_unlink ( name -- f ) ( 0 if okay, -1 if error )
  cstr sys_unlink syscall retval lnswap ldrop
;

: u_lseek ( whence d.byteno unix_fd --- d.byteno )
  >r l>r n->l lr> r> n->l  ( convert whence and unix_fd to longs )
  sys_lseek syscall lretval
  l>r ldrop ldrop ldrop lr> ( get rid of arguments )
;

: unixfseek ( d.byteno unix_fd --- )
    >r
    u.l_set lnover r>  ( d.byteno whence d.byteno unix_fd )
    u_lseek         ( d.byteno d.byteno' )
    l2dup l=
      if   ldrop ldrop   (  )
      else ." u_lseek failed - got " l. ." tried " l. quit then
;
: unixftell ( unix_fd -- d.current-position )
  >r u.l_incr 0. r> u_lseek
;
: unixfsize ( unix_fd --- d.size )
  >r
    ( remember the current position )
  r@ unixftell         ( d.current-position )
    ( seek to end of file to find out where the eof is )
  u.l_xtnd 0. r@ u_lseek ( d.current-position d.size )
    ( return to the original position )
  lswap r> unixfseek   ( d.size )
;
 
: unixfread ( bfsize bfbase l.byteno unix_fd -- #read l.aligned )
  lnswap                  ( bfsize bfbase fd l.byteno )
  unixalign               ( bfsize bfbase fd l.aligned )
  ldup l>r over unixfseek ( bfsize bfbase fd )           ( ret: l.aligned )
  u_read                  ( bytes_read )                 ( ret: l.aligned )
  dup -1 =  if  1 abort" unixfread failed"  then
  lr>                     ( #read l.aligned )
;

: unixfwrite ( bfend bftop bfbase fstart unix_fd  --- bfwrittend )
  >r
  r@  unixfseek                ( bfend bftop bfbase )
  swap over - over r@          ( bfend bfbase   bytecount addr unix_fd )
  u_write                      ( bfend bfbase   bytes_written )
  +                            ( bfend bfwrittenend )
  r> drop swap drop
;
 
: unixfclose ( fd --- )
  fid @ u_close
;

: nullread ( bfsize bfbase l.byteno unix_fd -- bfsize l.byteno )
  drop    ( don't need unix_fd )  ( bfsize bfbase l.byteno )
  l>r 2drop 0 lr>                 ( 0 l.byteno )
;
: nullwrite ( bfend bftop bfbase fstart fid --- )
  1 abort" You tried to write to a read-only file!"
; 

: setupfd ( mode unix_fd bufstart bufend fd )
  >r swap   r@   usebuf
  ( unix_fd )  dup r@ fid !      unixfsize   r@ fsize l!
  ( mode ) dup r@ fmode !
  ( mode ) dup read and
  if  ['] unixfread  else ['] nullread then ( mode readroutine )
  swap write and
  if  ['] unixfwrite else ['] nullwrite then ( readroutine writeroutine )
  ['] unixfclose ['] nodev
  r>  useroutines
;
 
: unixfcreate ( name --- f ) ( true success, false failure )
  file-protection @ swap ( protection name ) u_creat ( unix_fd )
  dup 0<
  if   drop  false
  else u_close true
  then ;
: unixfdelete ( name --- f ) ( 1 success, 0 failure )
  u_unlink  0=
;

: unixfopen ( name mode --- fd ) ( 0 if failure )
  dup >unixmode rot  ( mode   unixmode name )
  u_open ( mode unix_fd )
  dup 0<
  if   2drop 0   ( open failed )
  else find-fd dup (find-fbuf swap ( mode unix_fd buf fd )
       >r dup bufsize @ + r@ setupfd r> ( fd )
  then
;

create smallbuf  4 allot

: setup-stdio-fd ( mode unix_fd cfa -- )
  find-fd dup rot >body !   ( mode unix_fd fd )
  >r >r
  r@ ( mode unix_fd )
( The next line is a crock.  It uses the unix_fd as an index into the )
( smallbuf array.  However, the unix fds for stdin, stdout, and stderr )
( are not likely to ever change from 0, 1, and 2 )
    smallbuf r> +  dup 1+  r>  setupfd
;

0 constant stdin
0 constant stdout
0 constant stderr

: setup-stdio ( -- )
  read  0 ['] stdin  setup-stdio-fd
  write 1 ['] stdout setup-stdio-fd
  write 2 ['] stderr setup-stdio-fd
;
: (cold-hook
  (cold-hook
  init-malloc
  init-bufs
  init-fds
  clear-fds
  setup-stdio
;
' (cold-hook is cold-hook

: open ( name mode -- fd )
  depth 2 < abort" open needs 2 arguments - name and mode"
  unixfopen
;
: make ( name -- f )
  unixfcreate
;
: delete ( name -- f )
  unixfdelete
;
